(* On utilise Camlp4, donc:
 
   i) en mode interprete, on utilise 
 
       ocaml camlp4o.cma pa_extend.cmo q_MLast.cmo extend_monade.cmo

  ii) en mode compile, on utilise

       ocamlc -pp "camlp4o pa_extend.cmo q_MLast.cmo extend_monade.cmo" -I +camlp4 -c interp-monad-evolution.ml

 iii) comme preprocessing, on ecrit

       camlp4o pa_extend.cmo q_MLast.cmo ./extend_monade.cmo pr_o.cmo interp-monad-evolution.ml

 *)

(*                                                                      *)
(* Semantique Denotationnelle comme Interpretation de TFun              *)
(*                                                                      *)
(* Roberto Di Cosmo (roberto@dicosmo.org)            2004               *)
(*                                                                      *)
(* Syntaxe Abstraite du Langage                                         *)
(*                                                                      *)

type tfuntype  = TBool | TNat | Arr of tfuntype * tfuntype   (* les types       *)

type  id        = string                                      (* identificateurs *) 

type  op         = Plus
              | Minus
	      | Mult
	      | Div

type  compop     = Eq
              | Less
	      | More

type  exp       = INT of int | BOOL of bool                   (* les expressions *)
              | Iden of id
              | App of (exp * exp)
              | Abs of ((id * tfuntype) * exp)
              | IF of (exp * exp * exp)
              | Compop  of compop * exp * exp
              | Base of (op * exp * exp)
              | Rec of (id * id * exp)                        (* recursion      *)
              | Raise                                         (* error          *)
              | Lkp of exp                                    (* effets de bord *)
              | Ref of exp
              | Upd of exp * exp
              | Seq of exp * exp


(* les operation de base *)

let getop = function
    Plus -> fun x y -> x+y
  | Minus -> fun x y -> x-y
  | Mult -> fun x y -> x*y
  | Div -> fun x y -> x/y
;;

let getcompop = function
    Eq -> fun x y -> x=y
  | Less -> fun x y -> x<y
  | More -> fun x y -> x>y
;;


(*                                                                           *)
(*            TFun : l'Interpreteur                                          *)
(*                                                                           *)
            
(* Ce code est une traduction en ML immediate de la Semantique               *)
(* Denotationnelle Directe de TFun dans FTH qu'on a donne' en cours          *)
(* Un grande avantage de cet approche est la clarte du code,                 *)
(* qui se commente tout seul                                                 *)
(* On remarquera cependant que on n'utilise pas ici les types du langage au  *)
(* contraire de ce que l'on fait dans la semantique mathematique.            *)
(* En effet, pour simplicite, on n'utilise pas ici la derivation de typage   *)
(* pour la definition inductive, mais seulement la structure des programmes. *)
(* Cela ne change pas vraiement les fonctions semantiques, mais  on          *)
(* accepte comme parametre des programmes qui sont mal types, ce qui         *)
(* apparait en trois points: l'erreur possible dans projectvalue et les deux *)
(* match incomplets pour Arrow et Bool. En effet, sans derivation de typage, *)
(* on ne peut pas garantir que le programme passe en parametre fournisse les *)
(* bons valeurs dans ces cas                                                 *)            

(* monads and monad transformers *)

(* for each relevant computational type, we give here the associated monad and monad transformers *)


module type Monad = 
   sig
     type 'a t
     val tval: 'a -> 'a t
     val tlift: ('a -> 'b t) -> ('a t -> 'b t)
         (* to unpack the computation *)
     val reveal: 'a t -> 'a
   end

module Id =
  struct
    type 'a t = 'a
    let tval v = v
    let tlift f = f
    let reveal v = v
  end

module Exc =
  struct
    type 'a t = Res of 'a | Err
    let tval v = Res v
    let tlift f e = match e with Err -> Err | Res v -> f v
    let reveal lv = match lv with Err -> failwith "Error: cannot project Err" | Res v -> v

    let raise () = Err
  end

module State = 
  struct
    type loc = int
    type sv = int
    type s = loc -> sv 
    type 'a t = s -> 'a * s
    let tval (v: 'a) = (fun s -> (v,s) : 'a t)
    let tlift (f: 'a -> 'b t) (e: 'a t) = 
        ((fun s -> let (v,s') = e s 
                   in f v s') : 'b t)

    (* Uninitialised memory *)

    let emptys = (fun l -> failwith "State: access to uninitialised memory location")

    let reveal texp = let (v,s') = texp emptys in v

    (* newloc, update, lkp, init *)

    let newloc = let l = ref 0 in fun () -> let nl = !l in l:= nl+1; nl

    let update (l:int) (v:sv) s = let s' = fun l' -> if l = l' then v else s l' in ((),s')
    let lkp l = fun s -> ((s l:sv),s)
    let init (v:sv) = fun s -> let l = newloc() in 
                          let s' = fun l' -> if l = l' then v else s l' in (l, s')
  end

    
(**** La construction de l'interprete comme foncteur dependant d'une monade *)

module Interp (T: Monad) =
struct
(* Definition des Domaines Semantiques:                                      *)

type cpo = (* les valeurs de base: ici les entiers et booleens *)
           Int of int | Bool of bool 
           (* les constructions du domaine *)           
	 | Unit of unit
         | Arrow of (idt -> expt)
and  idt  = cpo                (* CBV *) 
and  expt = cpo T.t

(* ouverture de la definition de la monade *)

let (tval: idt -> expt) = T.tval

let tlift (f:idt -> expt) (texp: expt) = T.tlift f texp

(* ceci est un point fixe pour le CBV de ML *)

let rec fix f = f (fun x -> (fix f) x)

(* acceder a l'environnement *)

let rec proj (i,(env: (id * idt) list)) =
      match env with
        []        -> failwith "Identificateur non trouve"
      | (id,a)::r -> if id = i then a else proj (i, r)

(* la fonction semantique *)

let rec (esem: exp -> (id * idt) list -> expt) = fun  exp env ->
        match exp with
          App (exp1,exp2) -> let_T (Arrow f) <= esem exp1 env in                                         (* CBV *)
	                     let_T v <= esem exp2 env in
                              f v
        | Abs ((id,t),exp)-> tval(Arrow(fun a -> (esem exp ((id,a)::env))))
        | Iden(id)        -> tval(proj(id,env))
        | INT i           -> tval(Int i)
        | BOOL b          -> tval(Bool b)
        | IF(b,exp1,exp2) -> let_T Bool bv <= esem b env in
                             if bv then esem exp1 env
                             	   else esem exp2 env
        | Compop(op,exp1,exp2) -> let_T Int n <= esem exp1 env in
                                  let_T Int m <= esem exp2 env in
                                    if (getcompop op n m) then tval(Bool true)
                                                          else tval(Bool false)
        | Base(op,exp1,exp2) -> let_T Int n <= esem exp1 env in
                                let_T Int m <= esem exp2 env in
                                 tval (Int ((getop op) n m))
        | Rec(f,x,exp)    -> tval(Arrow(fix(fun g -> (fun a -> (esem exp ((x,a)::(f,Arrow g)::env))))))  (* CBV *)

end

(* les 3 interpretes *)

module IId = Interp Id;

module IExc = Interp Exc;;

module IState = Interp State;;

(* voila quelques programmes pour tester le noyau du langage *)

(* Programme bien type *)
let r1 = esem (App(Abs(("x",TBool),IF(Iden "x",Base(Plus,INT 1,INT 5),INT 2)),BOOL true)) [];;

(* Programme mal type *)
let r2 = esem (App(INT 1, INT 2))[];;

(* Factoriel *)

let r3 = esem (App(Rec("fact","x",IF(Compop(Eq,Iden "x",INT 0),INT 1,Base(Mult,Iden "x", App(Iden "fact",Base(Minus,Iden "x",INT 1))))),INT 10)) [];;




